"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const puppeteer_extra_1 = __importDefault(require("puppeteer-extra"));
const puppeteer_extra_plugin_stealth_1 = __importDefault(require("puppeteer-extra-plugin-stealth"));
const config_1 = __importDefault(require("./config"));
const selectors_1 = __importDefault(require("./selectors"));
const service_1 = __importDefault(require("./service"));
class MessagesClient extends events_1.EventEmitter {
    constructor(options = { headless: "new" }) {
        super();
        this.isAuthenticated = false;
        this.launch(options);
    }
    async launch(options) {
        const browser = await puppeteer_extra_1.default.use((0, puppeteer_extra_plugin_stealth_1.default)()).launch({
            headless: options.headless,
            // devtools: true,
            args: [
                "--disable-web-security",
                "--disable-features=IsolateOrigins,site-per-process",
                "--no-sandbox",
                "--disable-setuid-sandbox",
                "--disable-dev-shm-usage",
                "--disable-accelerated-2d-canvas",
                "--disable-gpu",
                "--window-size=1920x1080",
                "--start-maximized",
            ],
        });
        this.browser = browser;
        this.page = await browser.newPage();
        if (options.credentials) {
            await this.restoreSession(options.credentials);
            return;
        }
        await this.page.goto(config_1.default.urls.auth, { waitUntil: "load" });
        await this.checkRememberMe();
        await this.attachQrCodeListener();
        await this.attachAuthListener();
    }
    async restoreSession(credentials) {
        await this.page.goto(config_1.default.urls.auth, { waitUntil: "load" });
        await this.checkRememberMe();
        await this.page.setCookie(...credentials.cookies);
        await this.page.evaluate((localStorageValues, sessionStorageValues) => {
            Object.keys(localStorageValues).forEach((key) => {
                window.localStorage.setItem(key, localStorageValues[key]);
            });
            Object.keys(sessionStorageValues).forEach((key) => {
                window.sessionStorage.setItem(key, sessionStorageValues[key]);
            });
        }, credentials.localStorage, credentials.sessionStorage);
        await this.page.goto(config_1.default.urls.conversations, {
            waitUntil: "domcontentloaded",
        });
        await this.page.waitForXPath(selectors_1.default.xpath.conversationList, {
            visible: true,
        });
        this.isAuthenticated = true;
        console.log("Restored session!");
        const service = new service_1.default(this.page);
        this.emit("authenticated", service);
    }
    async attachAuthListener() {
        this.page.on("framenavigated", async (frame) => {
            const url = frame.url();
            if (url.includes("/conversations")) {
                console.log("Authenticated!");
                this.isAuthenticated = true;
                const credentials = {
                    cookies: await this.page.cookies(),
                    localStorage: await this.page.evaluate(() => {
                        const json = JSON.stringify(window.localStorage);
                        return JSON.parse(json);
                    }),
                    sessionStorage: await this.page.evaluate(() => {
                        const json = JSON.stringify(window.sessionStorage);
                        return JSON.parse(json);
                    }),
                };
                this.emit("credentials", credentials);
                const service = new service_1.default(this.page);
                this.emit("authenticated", service);
                this.page.removeAllListeners("framenavigated");
            }
        });
    }
    async checkRememberMe() {
        const rememberMeSlider = await this.page.waitForXPath(selectors_1.default.xpath.rememberMeSlider, {
            visible: true,
        });
        if (!rememberMeSlider) {
            throw new Error("Could not find remember me slider");
        }
        const sliderBtn = await this.page.waitForXPath(selectors_1.default.xpath.rmemberMeSliderButton);
        const className = (await (await sliderBtn?.getProperty("className"))?.jsonValue()) ?? "";
        const isChecked = className.includes("checked");
        if (!isChecked && sliderBtn) {
            await (await sliderBtn.toElement("button")).click();
            console.log("Checked remember me!");
        }
    }
    async attachQrCodeListener() {
        await this.page.exposeFunction("onQrCodeChange", (base64Image) => {
            console.log("QR Code changed!");
            console.log({ base64Image });
            this.emit("qr-code", base64Image);
        });
        await this.page.waitForXPath(selectors_1.default.xpath.qrCode, { visible: true });
        await this.page.evaluate((qrCodeXpath) => {
            function getElementByXPath(xpath) {
                const snapshots = document.evaluate(xpath, document, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
                const results = [];
                for (let i = 0; i < snapshots.snapshotLength; i++) {
                    results.push(snapshots.snapshotItem(i));
                }
                return results;
            }
            const observer = new MutationObserver((mutations) => {
                for (const mutation of mutations) {
                    console.log({ mutation });
                    mutation.addedNodes.forEach((node) => {
                        if (node instanceof HTMLImageElement) {
                            console.log("QR Code changed!");
                            node.onload = () => {
                                console.log({ src: node.src });
                                console.log({ src: node.getAttribute("src") });
                                window.onQrCodeChange(node.src);
                            };
                        }
                    });
                    //   if (mutation.attributeName === "data-qr-code") {
                    //     // const base64Image = mutation.target
                    //     // window.onQrCodeChange(base64Image)
                    //     console.log("QR Code changed!")
                    //     const qrElement = mutation.target as Element
                    //     const img = qrElement.querySelector("img")
                    //     console.log(img)
                    //     if (img) {
                    //       const base64Image = img.getAttribute("src")
                    //       window.onQrCodeChange(base64Image ?? "")
                    //     }
                    //   }
                }
            });
            const qrCode = getElementByXPath(qrCodeXpath)[0];
            console.log({ qrCode });
            if (qrCode) {
                observer.observe(qrCode, {
                    attributes: true,
                    childList: true,
                    characterData: true,
                });
            }
            return observer;
        }, selectors_1.default.xpath.qrCode);
    }
}
exports.default = MessagesClient;
